try:
    import re
    import os
    import socket
    import logging

    # Add special-case error handling for the scenario where the settings.yml
    # file is unavailable, since we cannot include this file in our
    # RPM package. In this case, we can set data to an empty dictionary,
    # since none of the hardware tests require this information.
    try:
        from config_parser import data
    except IOError as e:
        if 'DCI_ENV' not in os.environ:
            raise e
        data = dict()

    import steps.utilities
    steps.utilities.safe_selenium_imports(__file__, globals())
except ImportError as e:
    print("ImportError in {0}: {1}".format(__file__, e))
    raise e

def before_all(context):
    context.data = data

    # Determine the RHEL version of the system running Spicegate.
    # This will be included in all test result files
    # to simplify the review process.
    try:
        context.RHELVersion = steps.utilities.runShellCommand('cat /etc/redhat-release')[1]
    except (TypeError, IndexError):
        # No need to reraise the exception here, because this logging is not
        # functionally critical to test execution.
        print("Unable to determine the test system's RHEL version.")

    ###
    # Note (jkinsman: 01/24/2018)
    #      The Firefox instance used by Selenium is configured, instantiated,
    #      and destructed on a per-testcase basis. This makes perfect sense,
    #      given that some test scenarios use Selenium, while others do not.
    #
    #      Unfortunately, the version of Firefox used on the test system
    #      determines how Selenium needs to interact with Firefox.
    #      By default, Selenium uses its implementation of the
    #      Marionette protocol to interact with Firefox.
    #
    #      You can read about the Marionette protocol here:
    #      https://firefox-source-docs.mozilla.org/testing/marionette/marionette/index.html
    #
    #      Unfortunately, Selenium has deprecated its implementation of
    #      the Marionette protocol in favor of a W3C-compliant WebDriver
    #      implementation. As of Firefox 53, Marionette is the only
    #      way to control Firefox via IPC.
    #
    #      Since Selenium's new WebDriver implementation cannot
    #      communicate with Firefox directly, its requests must be run
    #      through a middleware layer that translates WebDriver commands
    #      to Marionette commands. This is done in one of two ways:
    #
    #      (1) Using Selenium's deprecated Marionette implementation
    #          (this is the default behavior, and it's completely broken
    #           on Firefox 53 and above)
    #      (2) Using a Geckodriver binary found on the system's $PATH.
    #          Geckodriver is a piece of WebDriver <=> Marionette middleware
    #          that is developed by Mozilla.
    #          It is free, open source, and written in Rust.
    #
    #      This logic will determine whether or not Geckodriver is required
    #      based upon the version of Firefox that is installed on the system.
    #      It will only attempt to enable Geckodriver support if the version
    #      of Firefox is sufficiently new (>= 47), and the geckodriver binary
    #      is accessible via the $PATH. Both of these properties will be true
    #      on all systems that run Spicegate tests as part of our CI
    #      framework.
    #
    #      You can find a concise explanation of the points I addressed above
    #      here:
    #      https://stackoverflow.com/questions/43272919
    ###
    capabilities = None
    try:
            capabilities = webdriver.common.desired_capabilities.DesiredCapabilities().FIREFOX.copy()
    except NameError as e:
        # If a NameError was thrown, the Selenium module could not be
        # properly imported by the steps.utilities.safeSeleniumImports()
        # method. We must skip web UI tests in this scenario.
        pass
    else:
        firefoxVersionRegex = 'Mozilla Firefox (\d+)\.[\d+\.]*\d'
        firefoxInfo = steps.utilities.runShellCommand('which firefox')
        geckodriverInfo = steps.utilities.runShellCommand('which geckodriver')

        # If both Firefox and Geckodriver are installed on the host, we need to
        # check the Firefox version to determine if Selenium's Marionette
        # implementation needs to be disabled.
        if firefoxInfo[0] == 0 and geckodriverInfo[0] == 0:
            try:
                firefoxVersionInfo = re.search(
                    firefoxVersionRegex,
                    steps.utilities.runShellCommand('firefox -v')[1]
                ).group(1)

                if int(firefoxVersionInfo) > 52:
                    print("Enabling Marionette support via Geckodriver.")
                    capabilities["marionette"] = True
                else:
                    print("Disabling Marionette support via Geckodriver.")
                    capabilities["marionette"] = False
            except AttributeError as e:
                print("{0}: {1}".format(e.__class__, e))
                print('Unable to properly parse Firefox version information.')
                print('Launching Firefox using default configuration.')

    # Cache the desired Firefox browser capabilities within the
    # context option, where they can be used to configure the browser session
    # within the before_scenario callback.
    context.capabilities = capabilities

    #   test_setup(context, "rhcert")


def before_scenario(context, scenario):
    # Store a refence to the 'scenario' object that represents
    # the scenario the Behave framework is currently executing.
    context.running_scenario = scenario

    # separator := string to visually separate important test information.
    separator = "-" * 75

    # Log some basic information about our test system
    # to simplify manually reviewing our test result files.
    # This functionality is in no way required, and is provided
    # as a "best-effort" service.
    try:
        rhcertPackages = steps.utilities.runShellCommand("rpm -qa 'redhat-certification*'")[1]
        print("\n".join([separator, "RHEL VERSION:", context.RHELVersion]))
        print("\n".join([separator, "REDHAT CERTIFICATION PACKAGES TESTED:", rhcertPackages, separator]))
    except (TypeError, IndexError, AttributeError):
        # No need to do anything here, since this functionality is
        # "best-effort", and these errors are not fatal.
        pass

    # Given that the majority of test cases leverage the Web UI,
    # it makes more sense to add conditional logic for tests that
    # do not use Selenium.
    if 'no_web_ui' in scenario.tags:
        return

    if context.capabilities is None:
        scenario.skip("Cannot execute web UI tests without Python's Selenium package.")
        return

    selenium_logger = logging.getLogger(
        'selenium.webdriver.remote.remote_connection')
    selenium_logger.setLevel(logging.WARN)
    profile = webdriver.FirefoxProfile()
    profile.set_preference("browser.cache.disk.enable", False)
    profile.set_preference("browser.cache.memory.enable", False)
    profile.set_preference("browser.cache.offline.enable", False)
    profile.set_preference("network.http.use-cache", False)

    context.browser = webdriver.Firefox(profile, capabilities=context.capabilities)
    # context.browser = webdriver.Firefox(capabilities=context.capabilities)
    context.browser.set_page_load_timeout(7200)
    context.browser.implicitly_wait(30)


def after_scenario(context, scenario):
    try:
        context.browser.quit()
    except (AttributeError, SyntaxError) as e:
        # This error is not fatal for tests that include the
        # "no_web_ui" tag, because the context's browser attribute
        # was never initialized in the first place.
        if 'no_web_ui' not in scenario.tags:
            print("{0}: {1}".format(e.__class__, e))
            raise e

def before_feature(context, feature):
    # Store a refence to the 'feature' object that represents
    # the feature associated with the scenario that the
    # Behave framework is currently executing.
    context.running_feature = feature

    # feature_name = get_feature_name(str(feature.name))
    #    if feature_name != "None":
    #        test_setup(context, "rhcert")
    #        test_setup(context, feature_name)

def after_feature(context, feature):
    print("Initiated teardown after feature %s"%feature.name)
    feature_name = get_feature_name(str(feature.name))
    print(feature_name)
    #    if feature_name != "None":
    #        destroy_box(context,"rhcert")
    #        destroy_box(context,feature_name)

def get_feature_name(feature_name):
    names_dict = {"Cloud Certification": "None",
                  "Neutron Certification": "None",
                  "Cinder Certification": "None",
                  "Sahara Certification": "None",
                  "Manila Certification": "None",
                  "Other Certification": "None",
                  "Management application": "None",
                  "Container Certification": "None",
                  "RHCERT common functionality": "None",
                  "CWE": "None",
                  "CCSP Sandbox Testing": "cloud",
                  "Hardware Certification Info Test": "None",
                  "Ironic Certification": "None",
                  "Other Certification": "None",
                  "Result Upload": "None",
                  "Updating Certification": "None",
                  "Sandbox Testing": "None",
                  "VNF Certification": "None",
                  "Cloud Certification CLI": "None",
                  "Cloud Certification CI": "None",
                  "Management Functionality": "None",
                  "Web UI Login Page": "None",
                  "Hardware Certification Memory Test": "None",
                  "Hardware Certification CPUScaling Test": "None",
                  "Hardware Certification Core Test": "None",
                  "Hardware Certification Profiler Test": "None",
                  "Hardware Certification Storage Test": "None",
                  "Hardware Certification Video Test": "None",
                  "Hardware Certification FV Memory Test": "None",
                  "Hardware Certification FV Core Test": "None",
                  "Hardware Certification wui CPUScaling Test": "None",
                  "Hardware Certification wui Core Test": "None",
                  "Hardware Certification wui Profiler Test": "None",
                  "Hardware Certification wui FV Memory Test": "None",
                  "Hardware Certification wui FV Core Test": "None",
                  "Hardware Certification wui Memory Test": "None",
                  "Hardware Certification ci Info Test": "None",
                  "Hardware Certification ci CPUScaling Test": "None",
                  "Hardware Certification ci Core Test": "None",
                  "Hardware Certification ci SATA Test": "None",
                  "Hardware Certification ci STORAGE Test": "None",
                  "Hardware Certification ci Profiler Test": "None",
                  "Hardware Certification ci FV Memory Test": "None",
                  "Hardware Certification ci FV Core Test": "None",
                  "Hardware Certification ci Memory Test": "None",
                  "Hardware Certification wui Storage SATA Test": "None",
                  "Hardware Certification wui Storage Test": "None",
                  "Hardware Certification wui Audio Test":"None",
                  "Sandbox for Hardware": "None" ,
                  "Rhcert Version Adding": "None",
                  "HWCERT Support Case and KBase Updates": "None",
                  "Management Functionality For OpenStack": "None",
                  "Create a Product for CCSP": "None",
                  "Create Vendor for CCSP and add Product against it":"None",
                  "Create Vendor for OSP": "None",
                  "Create Vendor for Hardware": "None",
                  "Clear all the logs and Set debug priority high": "None"}
    return(names_dict[feature_name])

def after_all(context):
    print("After all features executed")
#    destroy_box(context,"rhcert")
