# Copyright (c) 2017 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Jake Kinsman <jkinsman@redhat.com>

# Attempt to import the requisite Python modules.
try:
    import utilities
    import os
    import sys
    import shlex
    import select
    import subprocess
    import behave
    import time
    import fcntl
    import socket
    import time
    try:
        import selenium.common.exceptions
    except SyntaxError as e:
        # This error is non fatal on RHEL 6, which should only be used as
        # "CLIENTS" systems within a multi-host test environment.
        if "SERVERS" in os.environ and os.environ["SERVERS"].find(socket.gethostname()) == 0:
            utilities.printError(e, raiseError=True)
        else:
            utilities.printError(e, raiseError=False)

except ImportError as e:
    utilities.printError(e, raiseError=True)

# hostname := hostname of system currently under test
__hostname = utilities.runShellCommand('hostname')[1].strip()

###
# Note (jkinsman: 11/21/2017)
#      The two dictionaries below (__RHCertWUIViews and __RHCertWUIElements)
#      detail mappings between logical names for views and UI elements
#      (respectively) and unique HTML name attributes that identify
#      the elements these logical names refer to. These data structures
#      provide similar functionality to the "HTML_LOOKUPS" dictionary
#      in the 'utilities' submodule. We should consider merging and extending
#      these data structures so that we can easily identify HTML elements
#      by logical name based upon a set of one or more HTML attributes.   
###

# RHCertWUIViews := dictionary that maps logical page names
#                   (I.e., login, certification_index, etc.)
#                   to the redhat-certification URLs associated
#                   with those views.
__RHCertWUIViews = {
    'login': '{0}/login'.format(__hostname),
}

# RHCertWUIViews := dictionary that maps unique UI elements
#                   (I.e., 'login button', 'select product dropdown')
#                   to the HTML 'name' attribute of the associated
#                   UI element.
__RHCertWUIElements = {
    'login button': 'login',
    'cancel button': 'opt_out_login',
}

# This step will verify that rhcert-cli
# is not currently running on the test system.
@behave.given("rhcert-cli is not running")
def step_impl(context):
    # rhcertLockfile := Absolute path to the lock file created by rhcert
    rhcertLockfile = "/var/lock/subsys/rhcert"

    # Run the "ps ax" shell command, and capture its logging to stdout.
    psOutput = utilities.runShellCommand("ps ax")[1]

    # Parse the output of the "ps" command, and look for the PIDs
    # of running rhcert-cli processes.
    #
    # 'ps ax' Output:
    #
    #    PID TTY      STAT   TIME COMMAND
    #    1 ?        Ss     0:09 /usr/lib/systemd/systemd --switched-root ...
    #    2 ?        S      0:00 [kthreadd]
    #    3 ?        S      0:00 [ksoftirqd/0]
    #    ...
    #    26888 pts/3    TN     0:00 sudo rhcert-cli run --test info
    #
    # Result:
    #    ["26888"]
    for pid in map(
        lambda y: y.split(' ')[0],
        filter(
            lambda x: "rhcert-cli" in x and 'grep' not in x,
            psOutput.split("\n"))):
                assert utilities.runShellCommand("kill -SIGTERM {0}".format(pid))[0] == 0, "Could not kill rhcert-cli process {0}.".format(pid)

    # Now that we know rhcert-cli is not running on the host,
    # we can go ahead and attempt to remove the lock file (if it exists).
    if os.path.isfile(rhcertLockfile):
        try:
            t = os.remove(rhcertLockfile)
        except OSError as e:
            utilities.printError(e, raiseError=True)

@behave.given('the "{test}" test is planned')
def step_impl(context, test):
    planCmd = "rhcert-cli plan"
    testNames = None

    tests = utilities.runShellCommand(planCmd, rhcertCommand=True)

    if int(tests[0]) != 0:
        print("{0} did not complete successfully.".format(planCmd))
        assert False, "Unable to parse test plan information from rhcert-cli."

    # Else, the 'rhcert-cli plan' completed successfully.
    # Retrieve the list of planned tests by parsing the command's output.
    try:
        testNames = map(
            lambda test: test.split(' ')[0], tests[1].split('\n')
        )
    except (AttributeError,) as e:
        pass

    testPlanned = reduce(
        lambda x, y: x or y,
        map(
            lambda plannedTest: test in plannedTest,
            testNames
        ),
        False
    )

    if not testPlanned:
        context.running_scenario.skip("{0} was not planned on system {1}.".format(test, __hostname))

# This step will run a single redhat-certification-hardware test
# (specified by <test>) using either rhcert-cli or rhcert-wui,
# based upon the value of <harness>.
# Currently, only rhcert-cli is supported.
@behave.when('I run the "{test}" test via "{harness}"')
def step_impl(context, test, harness):
    context.execute_steps(u"given rhcert-cli is not running")
    context.execute_steps(u'given the "{0}" test is planned'.format(test))
    if (harness == "rhcert-cli"):
        utilities.runShellCommand(
            "rhcert-cli run --test {0}".format(test),
            rhcertCommand=True
        )
    elif (harness == "rhcert-ci"):
        utilities.runShellCommand(
            "rhcert-ci run --test {0}".format(test),
            rhcertCommand=True
        )


# This step is responsible for running the 'rhcert-cli print' command,
# and attaching the output and exit status of this command to the test
# context.
@behave.when('I print the results')
def step_impl(context):
    (exitStatus, stdout, stderr) = utilities.runShellCommand(
        "rhcert-cli print",
        rhcertCommand=True
    )

    context.results = {
        "testExitStatus": exitStatus,
        "testOutput": stdout,
        "testErrorOutput": stderr,
    }

# This step is responsible for ensuring that <test>
# completed with a result of <result>.
#
# This is determined by examining the result of <test> within the last
# test run logged by "rhcert-cli print". 
@behave.then('The "{test}" test result should be "{result}"')
def step_impl(context, test, result):
    # Retrieve the "rhcert-cli print" output stored in the context object.
    output = context.results['testOutput']
    mostRecentResult = None
    try:
        # Search the output of 'rhcert-cli print' for the desired test result.
        #
        # Input:
        #
        # Test Run 1
        # ---------------------------------------------------------------
        # suspend             ......                               - FAIL
        # self_check                                               - PASS
        # audio               ......                               -
        # audio               ......                               -
        # battery             ......                               -
        # lid                 ,,,,,,                               -
        # cdrom       sr0     ......                               -
        # dvd         sr0     ......                               -
        # expresscard                                              -
        # usb2                ......                               -
        # usb3                /devices/pci0000:00/0000:00:14.0/usb2-
        # kdump       nfs                                          -
        # kdump       local                                        -
        # WirelessN   wlp3s0  ......                               -
        # memory                                                   -
        # core                                                     -
        # cpuscaling                                               -
        # fv_core                                                  -
        # fv_memory                                                -
        # profiler                                                 -
        # storage     host0   ......                               -
        # video               ......                               -
        # info                                                     - REVIEW
        # sosreport                                                - PASS
        #
        # Output: 
        # <test>              ......                               - <RESULT>
        #
        # => <RESULT>
        mostRecentResult = filter(lambda x: x.strip().find(test) == 0, output.split("\n"))[-1]
    except IndexError as e:
        # An IndexError here indicates that the test was never planned.
        # This indicates a major test failure.
        utilities.printError(e, raiseError=True)

    # Assert that the parsed test result matches our expectation.
    assert result.upper() in mostRecentResult, "{0} test did not {1}.".format(test, result)

# This step is responsible for starting the redhat-certification
# web UI if it is not currently running.
@behave.given(u'The rhcert web ui is running')
def step_impl(context):
    # Check if rhcertd is currently running
    (exitStatus, stdout, stderr) = utilities.runShellCommand(
        "rhcertd status",
        rhcertCommand=True
    )

    # Start the rhcert daemon if the output and exit status of the
    # 'rhcertd status' command indicate that it is not currently running.
    if (("The rhcert daemon is NOT running" in stdout or
        "the rhcert listener is NOT running" in stdout) and
        exitStatus != 0):
            print("Starting redhat-certification web UI.")
            assert int(utilities.runShellCommand('rhcertd start', rhcertCommand=True)[0]) == 0, "Unable to start redhat-certification web UI."

    else:
        print("Restarting redhat-certification web UI.")
        assert int(utilities.runShellCommand('rhcertd restart', rhcertCommand=True)[0]) == 0, "Unable to restart redhat-certification web UI."

# This step is responsible for directing the Selenium web browser to access
# the redhat-certification URL detailed by <view>. The exact URL that will
# be visited is the value associated with the <view> key within the
# __RHCertWUIViews dictionary.
@behave.when(u'I go to the rhcert "{view}" page')
def step_impl(context, view):
    # Make sure the redhat-certification web UI is currently running
    # context.execute_steps(u'Given The rhcert web ui is running')

    # Ensure the view specified by the feature file maps
    # to a known URL within our __RHCertWUIViews dictionary. 
    assert view in __RHCertWUIViews.keys(), "Invalid redhat-certification WUI view specified."

    print("Browsing to {0}.".format(__RHCertWUIViews[view]))

    # Send an HTTP GET request to the resolved URL, and render the response
    # using Firefox, Geckodriver, and Selenium.
    context.browser.get('http://{0}'.format(__RHCertWUIViews[view]))

    # Save a screenshot of the page we have navigated to.
    # This is extremely useful for debugging Selenium issues
    # in headless environments.
    if 'SPICEGATE_DIR' in os.environ:
        context.browser.save_screenshot(
            "{0}/screenshots/{1}.png".format(
                os.environ['SPICEGATE_DIR'],
                view.replace(' ', '-')
            )
        )

# This step is responsible for verifying that a specific UI element
# (logically identified by <UIElement>) exists within the view
# currently being rendered by the browser. The element's name attribute
# can be determined based upon the value of __RHCertWUIElements[<UIElement>].
@behave.then(u'I should see the "{UIElement}"')
def step_impl(context, UIElement):
    # Ensure the UI element specified by the feature file maps
    # to a known HTML 'name' attribute within our __RHCertWUIElements
    # deictionary.
    assert UIElement in __RHCertWUIElements.keys(), "Invalid redhat-certification WUI UI element specified."

    # Attempt to locate the specified UI element within the view currently
    # being rendered by the browser. A NoSuchElementException will be raised
    # if the element does not exist.
    try:
        context.browser.find_element_by_name(__RHCertWUIElements[UIElement])
    except selenium.common.exceptions.NoSuchElementException as e:
        print("Error {0}: {1}".format(e.__class__, e))
        assert False, "Could not locate UI element {0} on URI {1}".format(UIElement, context.browser.current_url)
