# Copyright (c) 2017 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


# Attempt to import the requisite Python modules.
# Print a descriptive error message to the user
# if any of these dependencies cannot be resolved.
#
# collections.Callable works for Python2 and all Python version 3 variants.
#
try:
    from behave import given, when
    import os
    import sys
    import shlex
    import select
    import subprocess
    import time
    import fcntl
    import socket
    import re
    from collections import Callable
except ImportError as e:
    print("{0} failed to import one or more dependencies, as indicated below:\n{1}: {2}".format(__name__, e.__class__, e))
    sys.exit(1)

HTML_LOOKUPS = {
    "Log In": "_eventId_submit",
    "Product URL": "product-url",
    "Program": "program_chosen",
    "Red Hat Product": "rh_product_chosen",
    "Version": "rh_product_version_chosen",
    "Architecture": "arch_chosen",
    "Format": "format_chosen",
    "Build": "build",
    "Product Name": "product_option_chosen",
    "Vendor Name": "vendor_chosen",
    "Component": "component_chosen",
    "Submit": "button",
    "Continue": "button"
}

# This helper function checks if a /dev/<disk> is mounted and
# then returns True or False.
#
# Ideal Condition for isDiskMounted function is FALSE.
def isDiskMounted(diskDevice):
    isMountedCommand = "grep -w -o " + diskDevice + " /proc/mounts"
    mounted = runShellCommand(isMountedCommand)[1]
    if diskDevice in mounted:
        return True
    return False

# This helper function checks if a /dev/<disk> is a boot partition ("/boot")
# and then returns True or False.
#
# Ideal Condition for isDiskBootPart function is FALSE.
def isDiskBootPart(diskDevice):
    isBootCommand = 'lsblk --noheadings --paths --output=NAME,MOUNTPOINT ' + diskDevice
    bootDevice = runShellCommand(isBootCommand)[1]

    # Extra space is important for pattern match.
    diskDevicePattern = '^{0} '.format(diskDevice)
    bootPartPattern = '/boot$'
    patterns = [ diskDevicePattern , bootPartPattern]

    # NOTE: re.search executes twice, once for each pattern.
    #  So both MUST be true for this to work.
    for pattern in patterns:
        if not re.search(pattern, bootDevice):
            return False
        return True

# This helper function checks if a /dev/<disk> is of TYPE disk from
# command lsblk --noheadings --raw --paths --output=NAME,TYPE /dev/sda
#
# Ideal Condition for isDiskType function is FALSE
# False implies that what is displayed/reported from TYPE column is "lvm" or "part"
# True implies device from the TYPE column TYPE "disk" - Not an Ideal choice
def isDiskType(diskDevice):
    isDiskType = 'lsblk --noheadings --raw --paths --output=NAME,TYPE ' + diskDevice
    deviceType = runShellCommand(isDiskType)[1].strip()
    diskDevicePattern = '^{0} disk'.format(diskDevice)
    patterns = [ diskDevicePattern ]

    # NOTE: re.search executes twice, once for each pattern.
    #  So both patterns MUST be true for this to work.
    for pattern in patterns:
        if re.search(pattern, deviceType):
            return True
        return False

# This helper function is only to check if a /dev/<disk> has SWAP on it
# and then returns True or False.
#
# Ideal Condition for isDiskSwap function is TRUE.
#
# NOTE: Later on, you can still test a disk that does NOT have SWAP, this is just
#       the Ideal condition.
def isDiskSwap(diskDevice):
    isSwapCommand = 'lsblk --output=NAME,MOUNTPOINT ' + diskDevice
    swapDevice = runShellCommand(isSwapCommand)[1].split()
    if '[SWAP]' in swapDevice:
        return True
    return False

# This function will loop through all the presented choices for any of
# Storage tests and will return the ideal choice as the response.
# which will be written as the choice you wanted to the presented
# question of which disk choice to test?
#
# Your kilometers may vary on what "Ideal" means to you.
# Currently for a disk to end up in the idealChoice set the
# Ideal conditions in order is:
#
#                  Ideal
#                  Boolean
# Function         Returned
# -------------------------
# isDiskType       False
# isDiskMounted    False
# isDiskBootPart   False
# isDiskSwap       True
#
#
# Once a disk has been added to the notIdealSet it does NOT ever get
# removed from it.
#
# Assuming the choices of Disks has already been vetted by
# the certification test(s) at this point, no need to prove the disks exist.
#
# NOTE: Even at the end of the analysis we may still have NO Ideal disk(s).
#       So we will just return the original default response.
#
# Feel free to add more functions to zero in on the Ideal choice.
#
def selectDiskToTest(choiceTuple, default):
    # Initialize Set
    notIdealSet = set()
    idealChoiceSet = set()

    for diskItem in choiceTuple:
        diskDevice = '/dev/' + diskItem.strip()

        # Begin decision procress
        # Is Disk is Mounted?
        #    True  = NOT IDEAL Case
        #    False = IDEAL Case
        if isDiskMounted(diskDevice) is True:
            # True: NOT IDEAL Case (add to notIdealSet)
            notIdealSet.add(diskItem)

        # is Disk the Boot Partition?
        #    True  = NOT Ideal Condition
        #    False = Ideal Condition
        if isDiskBootPart(diskDevice) is True:
            # True: NOT IDEAL Case (add to notIdealSet)
            notIdealSet.add(diskItem)

        # Does Disk have SWAP?
        #    True  = Ideal Condition
        #    False = NOT Ideal Condition
        if isDiskSwap(diskDevice) is False:
            # False: NOT IDEAL case (add to notIdealSet)
            notIdealSet.add(diskItem)

        # Now see if there is only one disk choice left (the Ideal Choice!) by subtracting
        # the two python sets.
        choiceSet = set(choiceTuple)
        idealChoiceSet = choiceSet - notIdealSet

    # After subtraction decide what to do.
    #
    # There is nothing in the idealChoiceSet therefore since there was no
    # ideal disk found then return the original default response.
    # NOTE: Original response already has the "\n" appended to it. So do NOT add it here.
    if len(idealChoiceSet) == 0:
        print ("WARNING: Zero choices in ideal Set therefore returning original response\n")
        return default

    # So a little overloading here. If there
    # is only one choice, our job is done, but if multiple
    # choices exist, pick a random choice anyway.
    #
    # NOTE/Warning:
    # Because python set's have no indexing and are unordered the choice returned
    # in the case of MULTIPLE CHOICES within the idealChoiceSet the item returned
    # will be random! In adding an item to a python set the order of adding is
    # not guaranteed!
    return list(idealChoiceSet)[0]

# This function is responsible for printing a helpful error message
# when an exception <e> is thrown. It will reraise the exception if
# <raiseError> is true, causing the step to fail.
def printError(e, raiseError=False):
    print("{0}: {1}".format(e.__class__, str(e)))
    if raiseError:
        raise e

# This function will parse the choices presented by the Certification test from stdout.
# choices == disk choices presented by any Storage test. (I.e. sda, sda1, sda2, all)
#
# The goal is to make a better choice then replace the default response, only
# if necessary.
#
# Once the stdout has processed place all the choices into a read only
# tuple. Then pass this tuple to each function as necessary
# to perform some analysis and then bubble up the "desired" or Ideal disk choice as the
# response through this function so the choice can be written to the question being asked
# by the certification test.
#
# output  = Is the stdout buffer, so the presented choices can be extracted and acted upon.
# default = saved default response that was originally going to be used.
# *args   = Is any other arguments that could be passed.
# choiceDiskTuple = Contains the initial read only disk choices presented by the certification test
#
def answerDiskController(output, default, *args):

    # remove "\n" here, so we can consistently handle any response later
    default = default.strip()

    # Parse output for the given choices and store for
    # processing
    pattern = re.compile(r'^\d+. (\w+)')
    rbracket = "]"
    lbracket = "["
    apostro = "'"
    choiceDiskTuple = ()

    output = output.split('\n')

    # Use the list to initially get the choices but, throw it
    # away in favor of a tuple.
    choiceList = []

    for line in output:
        match = pattern.findall(line)
        if match:
            choiceList.append(str(match).strip(rbracket).strip(lbracket).strip(apostro))

    # Cast from a list to a Tuple so to force to be read only
    choiceDiskTuple = tuple(choiceList)

    # Now begin the process of selecting the "ideal" choice for disk to test.
    response = selectDiskToTest(choiceDiskTuple, default)

    # Now Do NOT forget to append the newline "\n" back in here to the response otherwise
    # the test will hang forever waiting for that newline to move on and begin the test.
    return response + '\n'

# This function is responsible for running a shell command (<command>),
# and returning its stdout, stderr, and exit status.
# For rhcert-related commands, this function is also responsible
# for detecting when the process has prompted for input, and providing
# a reasonable response to prevent the process from blocking on input.
def runShellCommand(command, rhcertCommand=False):
    shellCommand = command
    child = None

    # redhat-certification scripts for which this function
    # will be responsible for providing input (if necessary)
    rhcertExecutables = ["rhcert-cli", "rhcertd", "rhcert-ci"]

    stdout, stderr = "", ""

    # Data structure that details known prompts for the hwcert info test,
    # along with some instruction for how to satisfy them.
    # If a key of rhcertPrompts exists within an output string,
    # and all prerequisite strings (if specified) have been
    # previously printed within the output stream, then
    # provide the associated response.

    rhcertPrompts = {
        "lock file /var/lock/subsys/rhcert found": {
            "prerequisites": list(),
            "response": "yes\n",
        },
        "/var/rhcert/results.xml": {
            "prerequisites": list(),
            "response": "yes\n",
        },
        "Would you like to continue?": {
            "prerequisites": [
                "Warning: The hardware devices have changed.",
            ],
            "response": "yes\n",
        },
        "Would you like to install them now?": {
            "prerequisites": [
                "The following packages are required for testing:",
            ],
            "response": "yes\n",
        },
        "Would you like to re-try installing the missing packages?": {
            "prerequisites": [
                "The following packages are required for testing:"
            ],
            "response": "no\n",
        },
        "suspend?": {
            "prerequisites": [
                "This test will suspend the operating system.",
                "Please resume by pressing the power button after suspend has completed.",
            ],
            "response": "no\n",
        },
        "Does this system have a function key (Fn) to suspend the system to mem?": {
            "prerequisites": [
                "This test will suspend the operating system.",
                "Please resume by pressing the power button after suspend has completed.",
            ],
            "response": "no\n",
        },
        "Does this system have a function key (Fn) to suspend the system to disk?": {
            "prerequisites": [
                "This test will suspend the operating system.",
                "Please resume by pressing the power button after suspend has completed.",
            ],
            "response": "no\n",
        },
        "Please describe the need for these repos": {
            "prerequisites": list(),
            "response": "redhat-certification-ci\n",
        },
        "Do you want to proceed with the operation?": {
            "prerequisites": [
                "This host is not registered with any server.",
            ],
            "response": "yes\n",
        },
        "which disk would you like to test": {
            "prerequisites": [
                "For Disk Controller",
            ],
            "response": lambda output : answerDiskController(output, "all\n")
        },
        "Planning the fv_storage test. Do you want to assign a device of the SUT to the guest machine for testing the storage in the guest machine?": {
            "prerequisites": list(),
            "response": "No\n",
        },
        "This test system is not registered with any server(LTS).": {
            "prerequisites": list(),
            "response": "no\n",
        }
    }

    # If the shell command was provided to this function as a string,
    # use the shell's lexical parser to tokenize the input, and store
    # the tokens within a list.
    if isinstance(command, (str, unicode)):
        try:
            shellCommand = shlex.split(command)
        # Catching base class 'Exception' here because shlex does not
        # document which exception(s) are thrown for shlex#split.
        except Exception as e:
            printError(e, raiseError=True)

    # If __runShellCommand was called to run a script that is provided by
    # redhat-certification (i.e., rhcert-cli, rhcert-ci, rhcertd, etc.),
    # We need to modify the call in order to properly perform I/O
    # operations with the child process. Specifically, we need to:
    # (1) Invoke the script via python explicitly, so that the '-u' option
    #     can be used. The -u option forces the child process to not
    #     buffer its stdout and stderr streams, which is necessary to
    #     detect when the process is waiting for input via stdin.
    # (2) Provide the absolute path to the specified redhat-certification
    #     script. We need to specify the absolute path to the script
    #     because we are running it explicitly via Python. This means that
    #     the script's location within the filesystem will not be resolved
    #     automatically via the PATH environment variable when the shell
    #     command is evaluated.
    for executable in rhcertExecutables:
        if executable in shellCommand[0]:
            fullPath = runShellCommand(
                "which {0}".format(shellCommand[0])
            )[1].strip()
            shellCommand = ["python", "-u", fullPath] + shellCommand[1:]
    try:
        # Create the child process to run <command>.
        child = subprocess.Popen(
            shellCommand,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        # Allow the child process time to properly initialize.
        time.sleep(1)

        # If we're running an rhcert-cli command, we need to monitor
        # the process's stdout to determine if the process requires any
        # information from the test via stdin. We'll need to carefully
        # manage any I/O with the child process to ensure it does not hang.
        if rhcertCommand:
            # Set the pipe ends (file descriptors) connected to the child
            # process to non-blocking mode. This will force the parent to
            # immediately read any output the child has printed, even
            # if a full line of output is unavailable.
            # This is necessary for us to identify when the child
            # process is blocked waiting for input.
            fcntl.fcntl(child.stdout.fileno(), fcntl.F_SETFL, os.O_NONBLOCK)
            fcntl.fcntl(child.stderr.fileno(), fcntl.F_SETFL, os.O_NONBLOCK)

            # Default values for the exit status, stdout logging,
            # and stderr logging of the rhcert child process that will
            # be returned to the calling function.
            childExitStatus = None

            # Create a select() polling object. This works similarly to the
            # Unix select() system call, and will help us determine when the
            # child process has new output to process.
            selectInterface = select.poll()

            # Register the child process's stdout and stderr for output.
            selectInterface.register(child.stdout, select.POLLIN)
            selectInterface.register(child.stderr, select.POLLIN)

            # Begin polling the child process for output.
            while True:
                pollData = selectInterface.poll(1)
                # This list will store the channels whose select event
                # has a hang-up status. The child process has terminated
                # if its stdout and stderr channels enter the "POLLHUP"
                # state. 
                channelHUP = list()
                # Iterate through the channels that were polled to examine
                # any I/O events that may have occurred.
                for selectEvent in pollData:
                    # Check to see if the child has written any new data
                    # to stdout. The business logic here is a bit
                    # more complicated, because we need to determine
                    # whether or not the child process requires
                    # any additional information.
                    if selectEvent[0] == child.stdout.fileno() and selectEvent[1] == select.POLLIN:
                        # child.stdout.flush()
                        # Perform a non-blocking read on the child's stdout
                        # channel.
                        output = child.stdout.read()
                        # Log the output that was retrieved from the child
                        # process, and append it to our cumulative output
                        # string.
                        print(output.strip())
                        stdout = stdout + output

                        # Determine if the output from the child process
                        # contains any strings that are known to require
                        # user input. These strings are detailed by the
                        # top-level keys in the rhcertPrompts data structure.
                        for key in rhcertPrompts.keys():
                            if key in output:
                                # If a prompt is found, iterate through the
                                # prompt's entry within the rhcertPrompts
                                # data structure, and ensure that each of
                                # the contextual strings in the object's
                                # prerequisites list have also been found
                                # within the child's output stream.
                                # If a prompt and all of its contextual
                                # prerequisites have been found within the
                                # child's output stream, write the
                                # pre-defined response to the child process
                                # via stdin.
                                prerequisitesSatisfied = True
                                for prereq in rhcertPrompts[key]["prerequisites"]:
                                    if prereq not in stdout:
                                        prerequisitesSatisfied = False
                                        break
                                if prerequisitesSatisfied:
                                    if isinstance(rhcertPrompts[key]["response"], Callable):
                                        # send the output from stdout so it can be processed
                                        processedResponse = str(rhcertPrompts[key]["response"](output))
                                        child.stdin.write(processedResponse)
                                    else:
                                        child.stdin.write(rhcertPrompts[key]["response"])
                                    # break here - No need to scan rhcertPrompts any more if found what you were looking for
                                    break

                    # Else, determine if the child has written any new data to stderr.
                    elif selectEvent[0] == child.stderr.fileno() and selectEvent[1] == select.POLLIN:
                        # child.stderr.flush()
                        # Read the child's output on stderr,
                        # store the information locally,
                        # and log it appropriately.
                        output = child.stderr.read()
                        print(output.strip())
                        stderr = stderr + output

                    # Else, check to see if the select event
                    # has an event code equal to select.POLLHUP.
                    # This indicates that the child has likely terminated properly.
                    elif selectEvent[1] == select.POLLHUP:
                        channelHUP.append(selectEvent[0])

                # If select() returns a select.POLLHUP event for all
                # registered channels, we can safely assume that the process
                # terminated successfully. We can now return the
                # returncode, stdout, and stderr of the rhcert command
                # by breaking the while loop.
                if len(channelHUP) == len(pollData) and len(pollData) != 0:
                    break

            # Although the process has already terminated, we need to
            # explicitly call communicate() to ensure the exist status is properly set.
            child.communicate()

            return (child.returncode, stdout, stderr)

        # Else, the command specified by the calling function is not
        # associated with a redhat-certification utility.
        # We can go ahead and run this command normally.
        (stdout, stderr) = child.communicate()

        return (child.returncode, stdout, stderr)

    except Exception as e:
        printError(e, raiseError=True)

@given('I visit {url}')
def step_impl(context, url):
    context.browser.get(url)


@when(u'I select \'{value}\' from the \'{select}\' dropdown')
def step_impl(context, value, select):
    chosen = context.browser.find_element_by_id(HTML_LOOKUPS[select])
    chosen.find_element_by_css_selector("a").click()
    results = chosen.find_elements_by_css_selector(".chosen-results li")
    for result in results:
        if result.text == value:
            break
    result.click()


@when('I fill in \'{textbox}\' with \'{text}\'')
def step_impl(context, textbox, text):
    selected_textbox = context.browser.find_element_by_id(HTML_LOOKUPS[textbox])
    selected_textbox.clear()
    selected_textbox.send_keys(text)


@when('I click on \'{button}\'')
def step_impl(context, button):
    selected_button = context.browser.find_element_by_id(HTML_LOOKUPS[button])
    selected_button.click()

def safe_selenium_imports(filename, scope):
    def raiseImportError():
        raise ImportError("Unable to import Selenium on unsupported Python 2.6 version from file {0}.".format(filename))

    try:
        from selenium import webdriver
        from selenium import common
        from selenium.webdriver.support import expected_conditions as ec
        scope['webdriver'] = webdriver
        scope['common'] = common
        scope['webdriver'].support.expected_conditions = ec
    except SyntaxError as e:
        print("A SyntaxError was observed while importing")
        print("the python-selenium module within {0}.".format(__file__))
        print("{0}: {1}".format(e.__class__, e))
        print("This only occurs for EL6 (python-2.6), and is therefore")
        print("only a true error if we are assigned a 'SERVERS' role in Beaker.")

        # To maintain default behavior, we're going to go ahead and reraise
        # this exception if Spicegate is not running on Beaker systems
        # as part of our CI workflow.
        if "SERVERS" not in os.environ and "CLIENTS" not in os.environ:
            raiseImportError()

        # Else, reraise an exception if the system running Spicegate is
        # fulfilling a "SERVERS" role within a multi-host testing environment
        # within our CI workflow.
        if os.environ["SERVERS"].find(socket.gethostname()) != -1:
            raiseImportError()

    except ImportError as e:
        print("A {0} was encountered while importing selenium from {1}.".format(e.__class__, filename))
        print("{0}: {1}".format(e.__class__, e))
        raise e
