#!/usr/bin/env python
# Copyright (c) 2017 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Jake Kinsman
#
# This script is responsible for:
# (1) Installing the redhat-certification packages under test from
#     the rhcert Yum repository. The specific package versions of interest
#     are specified by the $REDHAT_CERTIFICATION, $REDHAT_CERTIFICATION_BACKEND,
#     $REDHAT_CERTIFICATION_HARDWARE, and $REDHAT_CERTIFICATION_HARDWARE_PREVIEW
#     environment variables.
#
# (2) Configuring redhat-certification to use the CWE environment specified
#     by the PHASE_URL Beaker job parameter. The script will default to "qa"
#     if a valid phase url is not provided.
#
# (3) Invoking the Spicegate framework to run our automated tests,
#     and saving the test results locally in jUnit format.
#
# (4) Merging jUnit-formatted XML documents to remove duplicate test entries.
# Import the requisite Python modules, and print a descriptive error message
# if any of these dependencies cannot be resolved.
#

try:
    import os
    import json
    import sys
    import re
    import subprocess
    import xvfbwrapper
    import socket
    import itertools
    import shlex
    import shutil
    import socket
    import time

    # Using an aliased import to reduce verbosity and improve readability.
    import xml.etree.ElementTree as ET

except ImportError as e:
    print "{0} could not continue because a fatal error was encountered.".format(__name__)
    print "{0}: {1}".format(e.__class__, str(e))
    sys.exit(1)

# __xvfb_width :=  width (in pixels) of the virtual X display buffer used
#                by Selenium to render the redhat-certification web UI.
# __xvfb_height := height (in pixels) of the virtual X display buffer used
#                by Selenium to render the redhat-certification web UI.
__xvfb_width = 1920
__xvfb_height = 1080

# __testResultRootDir := Absolute path to the root of the directory structure
#                        where test results are processed before they are
#                        exported to Jenkins and (eventually) Polarion.
__testResultRootDir = "/test_results"

# __multiHostCommSocket := An internet (AF_INET, A.K.A. TCP/IP) socket
#                          to communicate between the SUT and LTS
#                          to coordinate testing.
__multiHostCommSocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
__multiHostCommSocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

###
# Note (jkinsman: 04/11/2018)
#      While it is typically frowned upon to bind on all network interfaces,
#      I believe it is necessary in this scenario. The Beaker lab network
#      is a complicated environment, and seems to vary on a per-system
#      or per-lab basis. Essentially, it is nontrivial to determine
#      which IP addresses can be routed to externally, and which addresses
#      can be used for internal services, such as Kickstart, PXE, and iPMI.
###
#
# __LTSBindIP := IP Address that the LTS will bind() to for the purpose
#                of establishing a TCP connection with the SUT.
__LTSBindIP = "0.0.0.0"

# __LTSBindPort := Port that the LTS will bind() to for the purpose
#                  of establishing a TCP connection with the SUT.
__LTSBindPort = 8482

# __syncMessageLength := Length (in bytes) of control messages sent between
#                        the LTS and SUT to coordinate testing efforts.
__syncMessageLength = 4

# __enterCSMsg := message written from one test system to the other to
#                 acknowledge their request for critical section execution.
__enterCSMsg = "test"

# __exitCSMsg := message written from one test system to the other to
#                acknowledge the request for critical section execution
#                is now complete.
__exitCSMsg = "done"

# This function is responsible for determining which CWE instance
# we will be testing against (i.e., qa, stage, hydraqa, etc.).
# This will be determined by checking the PHASE_URL parameter of the
# submitted Beaker job, whose value will be defined as a shell
# environment variable. For now, this parameter will not be
# specified by Jenkins, and the function will fallback to returning "qa"
# as a default environment.
def __getPhase():
    try:
        return os.environ['PHASE_URL']
    except KeyError as e:
        # Return "qa" if a PHASE_URL was not specified as a Beaker job parameters.
        return "qa"

# Return the RHEL version of the system under test by parsing the /etc/redhat-release file
def __getRHELVersion():
    versionFile = "/etc/redhat-release"
    versionRegex = "Red Hat Enterprise Linux[ a-zA-Z]* release (\d.\d+) \(([a-zA-Z]+)\)"
    try:
        with open(versionFile, 'r') as fd:
            # Group 1 corresponds to the OS major, minor version (I.e., 7.4 or 6.8)
            return re.search(versionRegex, fd.read()).group(1)
    except AttributeError as e:
        # Return an empty string if the version could not be properly parsed.
        return ""

# This function is responsible for returning a tag that indicates
# a subset of tests this system is responsible for executing based
# upon its role, and whether or not multihost testing needs to
# be conducted.
def __getSupplementalTag(role, multihost=False):
    if role == "SERVERS":
        return "multihost_lts_initiated" if multihost else "lts_only"
    return "multihost_sut_initiated" if multihost else "sut_only"

# (1) Installing the redhat-certification packages under test from
#     the rhcert Yum repository. The specific package versions of interest
#     are specified by the $REDHAT_CERTIFICATION,
#     $REDHAT_CERTIFICATION_BACKEND, $REDHAT_CERTIFICATION_HARDWARE, and
#     $REDHAT_CERTIFICATION_HARDWARE_PREVIEW environment variables.
def __installRequisistePackages(version):
    # This helper function is responsible for assembling the list of
    # packages that need to be installed on the system under test.
    def getPackageList():
        # Attempt to parse the specific package version(s) we need to test
        # from the environment variables defined by the Beaker job parameters.
        # Please note that exactly one of these variables
        # will be defined with non-zero length.
        try:
            rhcert = os.environ['REDHAT_CERTIFICATION']
            rhcertBackend = os.environ['REDHAT_CERTIFICATION_BACKEND']
            hwcert = os.environ['REDHAT_CERTIFICATION_HARDWARE']
            hwcertPreview = os.environ['REDHAT_CERTIFICATION_HARDWARE_PREVIEW']
        # If a KeyError is raised, the Beaker job was not submitted with the
        # necessary parameters. Attempt to return a sane default value
        # (i.e., the latest version) for each of the packages
        # that we will be testing.
        except KeyError as e:
            try:
                return ["redhat-certification-backend" if float(version) < 7.0 else "redhat-certification", "redhat-certification-hardware", "redhat-certification-hardware-preview"]
            # An Exception here indicates a nonrecoverable failure.
            except (TypeError, ValueError) as e:
                print "RHEL version could not be parsed successfully."
                return 1

        # ret := list of packages to be installed.
        ret = list()
        # Flag to determine whether or not a new hardware-certification
        # package triggered the automated test execution.
        hardwareRelease = False

        # If hwcert has non-zero length, a specific NVR was specified
        # as a Beaker job parameter for the redhat-certification-hardware
        # package. We must be sure to test this specific version.
        if len(hwcert):
            hardwareRelease = True
            ret = [hwcert, 'redhat-certification-hardware-preview']

        # If hwcertPreview has non-zero length, a specific NVR was specified
        # as a Beaker job parameter for the
        # redhat-certification-hardware-preview package.
        # We must be sure to test this specific version.
        elif len(hwcertPreview):
            hardwareRelease = True
            ret = ['redhat-certification-hardware', hwcertPreview]

        # If a new hardware-certification package triggered the automated test execution,
        # make sure to install the newly-built hardware packages, alongside the most recent
        # redhat-certification[-backend] release for the system under test.
        if hardwareRelease:
            return ["redhat-certification-backend"] + ret if float(version) < 7.0 else ["redhat-certification-backend", "redhat-certification"] + ret

        # Else, a redhat-certification build triggered the automated test execution.
        # Install the recently-built redhat-certification[-backend] package,
        # along with the most recent redhat-certification-hardware and
        # redhat-certification-hardware-preview releases
        # for the system under test.
        if len(rhcert):
            return [rhcert, "redhat-certification-hardware", "redhat-certification-hardware-preview"]
        if len(rhcertBackend):
            
            if float(version) < 7.0:
                return [rhcertBackend, "redhat-certification-hardware", "redhat-certification-hardware-preview"]
            return [rhcertBackend, "redhat-certification", "redhat-certification-hardware", "redhat-certification-hardware-preview"]

        return ["redhat-certification-backend", "redhat-certification", "redhat-certification-hardware", "redhat-certification-hardware-preview"]

    # packages ;= List of packages to be installed on the system under test.
    packages = getPackageList()

    # Install the packages under test, and wait for the process to exit gracefully.
    yum = subprocess.Popen(["yum", "install", "-y"] + packages)
    yum.communicate()

    # Return the Yum command's exit status.
    return yum.returncode

# (2) Configuring redhat-certification to use the CWE environment specified
#     by the PHASE_URL Beaker job parameter. The script will default to "qa"
#     if a valid phase url is not provided.
def __startRHCertd(phase):
    rhcertXML = None
    xmlRoot = None

    # Start rhcertd to produce the '/etc/rhcert.xml' file
    # that we need to configure.
    rhcertd = subprocess.Popen(['rhcertd', 'start'])
    # Wait for the daemon's parent process to exit gracefully.
    rhcertd.communicate()

    # Ensure rhcertd was started successfully.
    if rhcertd.returncode != 0:
        print "Unable to start rhcertd."
        return 1

    try:
        # Attempt to parse the contents of the /etc/rhcert.xml file,
        with open('/etc/rhcert.xml', 'r+') as fd:
            rhcertXML = ET.ElementTree(ET.fromstring(fd.read()))

            # Reset the file pointer so that we can rewrite its contents later.
            fd.seek(0)

            # xmlRoot := reference to the root element of the XML document.
            xmlRoot = rhcertXML.getroot()

            # If the root element (rhcert-environment) has a child whose tag
            # is "urls", set its phase attribute
            # to the value specified by <phase>.
            # This instructs rhcertd to communicate with the desired
            # CWE instance.
            xmlRoot.find('urls').set("phase", phase)

            # Write the XML changes back to the /etc/rhcert.xml file.
            rhcertXML.write(fd)

            fd.truncate()
            fd.close()

    except (AttributeError, OSError) as e:
        # Print an error message and return a non-zero value
        # if something goes wrong.
        print "{0}: {1}".format(e.__class__, e)
        print "Unable to set CWE phase to {0}".format(phase)
        return 1

    # Restart rhcertd, and return a non-zero exit status
    # if the daemon could not be restarted.
    rhcertd = subprocess.Popen(['rhcertd', 'restart'])
    rhcertd.communicate()

    if rhcertd.returncode != 0:
        print "Unable to restart rhcertd."
        return 1

    return 0

# This helper function is responsible for determining which role the system
# has been designated within a multihost environment.
def __getRole():
    if "SERVERS" in os.environ and os.environ["SERVERS"].find(socket.gethostname()) == 0:
        return "SERVERS"
    return "CLIENTS"

# This is a helper function that is responsible for determining which feature
# file(s) should be included in the Behave invocation for test discovery
# purposes. This function (along with __mergeTestResults) is responsible
# for preserving test data across multiple behave invocations, and limiting
# unnecessary test logging.
def __getRelevantFeatureFiles(desiredTags):
    # tagPermutations := List of all permutations of the tags we have
    #                    specified to determine which tests we would
    #                    like to run.
    tagPermutations = [tag for tag in itertools.permutations(desiredTags)]

    # tagRegex := Perl-style regex used to match all desired permutations
    #             of our testing tags.
    tagRegex = "|".join(map(lambda x: ".*".join(x).strip(), tagPermutations))
    ###
    # Note (jkinsman: 04/05/2018)
    #      While this syntax for invoking a shell command and capturing its standard output is
    #      archaic and not particularly readable, I had a lot of difficulty getting the subprocess
    #      module to properly parse this shell command. Even with the help of shell=True and
    #      the shell's lexical parser, the subprocess module simply would not
    #      tokenize the command in a predictable manner
    #      before forking the child process.
    #
    #      Backticks are exactly what you need for commands like this. :)
    ###
    return "|".join(
        filter(len, os.popen("find . -name \"*.feature\" -exec grep -siIE '%s' {} \; -print | grep -v @redhat_certification_ci" % (tagRegex)).read().split("\n"))
    ).strip()

# This helper function is responsible for converting a list of test tags
# into an absolute path within the __testResultRootDir directory hierarchy.
def __getTestResultDir(tags):
    return "{0}/{1}".format(__testResultRootDir, "_".join(tags))

# (3) Invoking the Spicegate framework to run our automated tests,
#     and saving the test results locally in jUnit format.
def __runTests(multihost=False):
    # Determine if the system has a role within a multi-host test environment
    # as either a client or server.
    role = __getRole()

    # If the system is acting as a server, create an X virtual frame
    # buffer to render the redhat-certification web UI in a headless
    # environment.
    if role == "SERVERS":
        xvfb = xvfbwrapper.Xvfb(width=__xvfb_width, height=__xvfb_height)
        xvfb.start()

    # testTag := Specific type of subtest we would like to execute,
    #            i.e., "sut_only" or "multihost_lts_initiated"
    testTags = [
        "redhat_certification_ci",
        __getSupplementalTag(
            role,
            multihost
        )
    ]
    # featureFiles := Feature files (whitespace-separated string) that indicates
    #                 which feature files contain one or more scenarios
    #                 that include our desired test tags.
    featureFiles = __getRelevantFeatureFiles(testTags)

    # If the system will be acting as a server within a multi-host
    # test environment, we need to ensure that it runs the single-host tests
    # that are tagged 'lts_only'. Otherwise, we can assume the system is a
    # client, and needs to run the tests that are tagged 'sut_only'.
    args = "--junit-directory {0} --junit".format(__getTestResultDir(testTags))

    # Append the list of test tags we're interested in exercising to behave's
    # command-line arguments.
    args += " ".join([" --tags {0}".format(t.strip()) for t in testTags])

    args += " -i '{0}'".format(featureFiles)
    
    # hwcertTestCommand := Shell command used to invoke behave in order
    #                      to execute the desired tests.
    hwcertTestCommand = "behave {0}".format(args)

    # This function is a parameterized decorator for our doTest function,
    # which is responsible for actually invoking our tests.
    # This function is used to change the test execution behavior,
    # based upon one of three scenarios:
    # (1) The automated tests are running in a single-host configuration:
    #     The doTest() appropriately manages test execution in this scenario.
    # (2) The auromated tests are running multihost tests as an LTS:
    #     The script will establish a TCP connection with the SUT,
    #     and send control messages to ensure mutually-exclusive execution for tests that require execution from both systems.
    # (3) The automated tests are running multihost tests as an SUT:
    #     The script will establish a TCP connection with the LTS,
    #     and allow the LTS to run its multihost tests before it begins
    #     to execute its own multihost tests (if applicable).
    def condMutuallyExclusive(multihost, role):
        # Non-parameterized function that decorates doTest.
        # It leverages closure to access the 'multihost' and 'role'
        # values that were initially passed from the calling scope. 
        def _ret(function):
            # Single-host testing configuration: No additional synchronization logic required
            def identity(*args, **kwargs):
                return function(*args, **kwargs)

            # Multi-host SUT configuration: connect() to the LTS, and allow it to execute its multi-host
            # tests before we begin to execute our own multi-host tests.
            def mutexClient(*args, **kwargs):
                while True:
                    try:
                        serverFQDN = os.environ['SERVERS']
                        __multiHostCommSocket.connect((serverFQDN, __LTSBindPort))
                    except (OSError, socket.error, socket.herror) as e:
                        print "{0}: {1}".format(e.__class__, e)
                        time.sleep(10)
                    except KeyError as e:
                        print "{0}: {1}".format(e.__class__, e)
                        return 1
                    else:
                        break

                # Else, the LTS has successfully connected to the SUT.
                # Send an acknowledgement that the LTS may proceed, and
                # wait for a response that indicates that testing is complete.
                if __multiHostCommSocket.send(__enterCSMsg) != __syncMessageLength:
                    print "Unable to properly send request to enter critical section."
                    return 1

                # Wait for the LTS to inform us that its testing has completed.
                while True:
                    LTSResponse = __multiHostCommSocket.recv(__syncMessageLength)
                    if LTSResponse == __exitCSMsg:
                        break

                # Invoke multi-host tests that are intended to be initiated by the SUT.
                ret = function(*args, **kwargs)
                # Inform the LTS that our multi-host testing has completed.
                __multiHostCommSocket.send(__exitCSMsg)

                return ret

            # Multi-host LTS configuration: accept() a connection
            # from the SUT, and execute all multi-host tests before
            # notifying the SUT that it may commence multi-host testing. 
            def mutexServer(*args, **kwargs):
                # Default values to describe the origin of the
                # TCP connection we will be accepting from the SUt. 
                SUTSocket, addr = [None, None]
                try:
                    # Bind, listen, and accept on 0.0.0.0:8482,
                    # which is not a well-known port.
                    __multiHostCommSocket.bind((__LTSBindIP, __LTSBindPort))
                    __multiHostCommSocket.listen(1)
                    SUTSocket, addr = __multiHostCommSocket.accept()

                    # Wait for the SUT to inform us that we may safely proceed with multi-host testing.
                    while True:
                        SUTResponse = SUTSocket.recv(__syncMessageLength)
                        if SUTResponse == __enterCSMsg:
                            break

                except (OSError, socket.error, socket.herror) as e:
                    print "{0}: {1}".format(e.__class__, e)
                    print "__multiHostCommSocket = {0}".format(__multiHostCommSocket)
                    try:
                        SUTSocket.close()
                    except AttributeError as _:
                        pass

                    return 1
                # Invoke multi-host tests that are intended to be initiated by the LTS.
                ret = function(*args, **kwargs)

                # Notify the SUT that it is now safe to proceed with multi-host testing.
                if SUTSocket.send(__exitCSMsg) != __syncMessageLength:
                    print "Unable to properly send request to enter critical section."
                    return 1

                try:
                    # Wait for the SUT to notify us that multi-host testing is complete.
                    while True:
                        SUTResponse = SUTSocket.recv(__syncMessageLength)
                        if SUTResponse == __exitCSMsg:
                            break

                except (OSError, socket.error, socket.herror) as e:
                    print "Unable to parse control message from SUT."
                    print "{0}: {1}".format(e.__class__, e)
                    return 1

                finally:
                    try:
                        # Attempt Close the connection to the SUT.
                        # This will throw an AttributeError()
                        # if the connection was never established.
                        SUTSocket.close()
                    except AttributeError as _:
                        pass

                return ret

            # Return value of the _ret function: Return the appropriate
            # wrapper function (defined above), based upon the type of testing we would like to perform.
            if multihost is True:
                if role is "SERVERS":
                    return mutexServer
                return mutexClient

            # Else, something went wrong. Return the identity function in hopes
            # of preserving valuable debbuging information.
            return identity

        # Return the non-parameterized decorater that was constructed via closure.  
        return _ret

    # This function is responsible for actually running the tests. :)
    @condMutuallyExclusive(multihost, role)
    def doTest():
        # Run hwcertTestCommand as a child process,
        # and return its exit status upon termination.
        try:
            testRun = subprocess.Popen(shlex.split(hwcertTestCommand))
            testRun.communicate()
        except (OSError, ValueError) as e:
            print "Command {0} failed.".format(hwcertTestCommand)
            print "{0}: {1}".format(e.__class__, e)
            return 1
        return 0

    ret = doTest()

    # Stop (and free) the X virtual frame buffer (if applicable).
    if role == "SERVERS":
        xvfb.stop()

    # Return the exit status of the doTest()
    # function (nonzero exit status for failure).
    return ret

# This function is responsible for invoking all single-host tests,
# which are determined based upon the system's multi-host role. 
def __runSingleHostTests():
    return __runTests()

# This function is responsible for invoking all multi-host tests,
# which are also determined based upon the system's role. 
def __runMultiHostTests():
    return __runTests(multihost=True)


# This function is responsible for consolidating two (or more) of Behave's
# jUnit-formatted XML test result files that have the same file name.
# This allows us to easily maintain unique test results
# for single-host and multi-host test runs. 
def __mergeFiles(role, resultFileName, *resultFiles):

    resultXML = None
    for i, resultFile in enumerate(resultFiles):
        # Parse the ith result file to be merged
        xmlData = ET.parse(resultFile)
        # Store a reference to the root of the XML document.
        xmlRoot = xmlData.getroot()
        xmlTests = list(xmlRoot)

        # Merge testsuite information from multiple test runs.
        # I.e., If the tests took 20 minutes on system 1, and 10 minutes
        # on system 2, update the element's 'time' attribute to 30 minutes.
        if i == 0:
            resultXML = ET.Element(xmlRoot.tag, xmlRoot.attrib)
        else:
            try:
                resultXML.attrib['tests'] = str(max([int(resultXML.attrib['tests']), int(xmlRoot.attrib['tests'])]))
                resultXML.attrib['time'] = str(float(resultXML.attrib['time']) + float(xmlRoot.attrib['time']))
                resultXML.attrib['failures'] = str(int(resultXML.attrib['failures']) + int(xmlRoot.attrib['failures']))
                resultXML.attrib['errors'] = str(int(resultXML.attrib['errors']) + int(xmlRoot.attrib['errors']))
                resultXML.attrib['skipped'] = str(int(resultXML.attrib['skipped'])  - (int(xmlRoot.attrib['tests']) - int(xmlRoot.attrib['skipped'])))
            except KeyError as e:
                print "Unable to merge jUnit XML files generated by Behave."
                print "{0}: {1}".format(e.__class__, e)
                return 1

        # Attempt to merge individual test cases on the basis of feature
        # name, and add them to the final result document. 
        for rawTestResult in xmlRoot:
            foundMatch = False

            for j, processedTestResult in enumerate(resultXML):
                if rawTestResult.attrib['name'] == processedTestResult.attrib['name']:
                    foundMatch = True

                    try:
                        if processedTestResult.attrib['status'] == 'skipped' and rawTestResult.attrib['status'] != 'skipped':
                            resultXML.remove(processedTestResult)
                            resultXML.insert(j, rawTestResult)
                            break

                    except KeyError as e:
                        print "Unable to merge jUnit test results: Unable to determine test case results."
                        return 1

            if foundMatch is False:
                resultXML.insert(0, rawTestResult)

    # Serialize the data structure that represents the processed test results
    # to a local file that can be uploaded to Jenkins.
    with open("{0}/{1}-{2}".format(__testResultRootDir, role, resultFileName), 'w+') as outputFile:
        outputFile.write(ET.tostring(resultXML, encoding='utf8', method='xml'))

    return 0

# (4) Merging jUnit-formatted XML documents to remove duplicate test entries.
def __mergeTestResults():
    role = __getRole()

    singleHostTags = ["redhat_certification_ci", __getSupplementalTag(role)]
    multiHostTags = ["redhat_certification_ci", __getSupplementalTag(role, multihost=True)]

    testSubDirs = [
        __getTestResultDir(singleHostTags),
        __getTestResultDir(multiHostTags)
    ]

    singleHostNames, multiHostNames = [list(), list()]

    try:
        # singleHostNames := List of file names containing single-host test results.
        singleHostNames = os.listdir(testSubDirs[0])
        # multiHostNames := List of file names containing multi-host test results.
        multiHostNames = os.listdir(testSubDirs[1])
    except OSError as e:
        pass

    # Aggregate all test results in a single directory. Dispatch result files
    # that need to be processed to the __mergeFiles function. 
    try:
        for resultFile in set(singleHostNames + multiHostNames):
            if resultFile in singleHostNames and resultFile not in multiHostNames:
                shutil.copy(
                    "{0}/{1}".format(testSubDirs[0], resultFile),
                    "{0}/{1}-{2}".format(
                        __testResultRootDir,
                        role,
                        resultFile
                    )
                )

            elif resultFile in multiHostNames and resultFile not in singleHostNames:
                shutil.copy(
                    "{0}/{1}".format(testSubDirs[1], resultFile),
                    "{0}/{1}-{2}".format(
                        __testResultRootDir,
                        role,
                        resultFile
                    )
                )

            else:
                __mergeFiles(role, resultFile, *["{0}/{1}".format(subDir, resultFile) for subDir in testSubDirs])

    # Print a descriptive error message if anything goes wrong here.
    except Exception as e:
        print "{0}: {1}".format(e.__class__, e)
        return 1

    return 0

def __main():
    ret = 0

    # (1) Installing the redhat-certification packages under test from
    #     the rhcert Yum repository. The specific package versions of interest
    #     are specified by the $REDHAT_CERTIFICATION, $REDHAT_CERTIFICATION_BACKEND,
    #     $REDHAT_CERTIFICATION_HARDWARE, and $REDHAT_CERTIFICATION_HARDWARE_PREVIEW
    #     environment variables.
    if __installRequisistePackages(__getRHELVersion()):
        return 1

    # (2) Configuring redhat-certification to use the CWE environment specified
    #     by the PHASE_URL Beaker job parameter. The script will default to "qa"
    #     if a valid phase url is not provided.
    if  __startRHCertd(__getPhase()):
        return 1

    # (3) Invoking the Spicegate framework to run our automated tests,
    #     and saving the test results locally in jUnit format.
    ret = ret | __runSingleHostTests()

    # Comment out multi-host test invocation until work for CERTPX-2214#
    # is complete.
    # ret = ret | __runMultiHostTests()

    # (4) Merging jUnit-formatted XML documents to remove duplicate test entries.
    ret =  ret | __mergeTestResults()

    __multiHostCommSocket.close()

    return ret

if __name__ == '__main__':
    sys.exit(__main())
